import { normalize } from '@angular-devkit/core';
import { strings } from '@angular-devkit/core';

import { apply, move, Rule, template, url, mergeWith, chain, Tree } from '@angular-devkit/schematics';
import { GeneratorInterface } from '../GeneratorInterface';
import { GeneratorHelper } from '../GeneratorHelper';

const API_SERVICE_FILE_PATH = 'src/app/api';
const API_MODULE_FILE_PATH = 'src/app/api';
const ROOT_REDUCER_FILE_PATH = 'src/app/store/reducers.ts';
const ROOT_SAGA_FILE_PATH = 'src/app/store/sagas.ts';

const getApiModulePath = function(options: any) {
  const module = options['apifolderpath'] === undefined ? 'api.module.ts' : `${options['apifolderpath']}/${options['apifolderpath']}.module.ts`;
  return `${API_MODULE_FILE_PATH}/${module}`;
};

const getApiServicePath = function(options: any) {
  const module = options['apifolderpath'] === undefined ? 'api.service.ts' : `${options['apifolderpath']}/${options['apifolderpath']}.service.ts`;
  return `${API_SERVICE_FILE_PATH}/${module}`;
};

const getApiFolderPath = function(options: any) {
  const apifolderpath = options['apifolderpath'] === undefined ? '' : options['apifolderpath'];
  return `${options.sourceDir}/app/api/${apifolderpath}`;
};

const getComponentPath = function(options: any) {
  const component = options['component'] === undefined ? '' : options['component'];
  return `${options.sourceDir}/app/components/${component}`;
};

export class AdminUserComponentGenerator implements GeneratorInterface {
  addDeclarationToApiService(options: any): Rule {
    return (tree: Tree) => {
      let readVal = tree.read(getApiServicePath(options));
      const declarationRecorder = tree.beginUpdate(getApiServicePath(options));
      if (readVal) {
        let val = readVal || '';
        if (val) {
          let str = val.toString();
          declarationRecorder.insertLeft(
            0,
            `import { ${strings.classify(options.name)}Service } from './${strings.dasherize(options.name)}/${strings.dasherize(options.name)}.service';\n`
          );
          if (str.indexOf('constructor(') > -1) {
            declarationRecorder.insertRight(
              str.indexOf('constructor(') + 'constructor('.length,
              `public ${strings.camelize(options.name)}: ${strings.classify(options.name)}Service, `
            );
          }
        }
      }
      tree.commitUpdate(declarationRecorder);
    };
  }

  addDeclarationToApiModule(options: any): Rule {
    return (tree: Tree) => {
      let readVal = tree.read(getApiModulePath(options));
      const declarationRecorder = tree.beginUpdate(getApiModulePath(options));
      if (readVal) {
        let val = readVal || '';
        if (val) {
          let str = val.toString();
          declarationRecorder.insertLeft(
            0,
            `import { ${strings.classify(options.name)}Service } from './${strings.dasherize(options.name)}/${strings.dasherize(options.name)}.service';\n`
          );
          if (str.indexOf('providers: [') > -1) {
            declarationRecorder.insertRight(str.indexOf('providers: [') + 'providers: ['.length, `${strings.classify(options.name)}Service, `);
          }
        }
      }
      tree.commitUpdate(declarationRecorder);
    };
  }

  addDeclarationToReducer(options: any): Rule {
    return (tree: Tree) => {
      let declarationRecorder, readVal;
      if (options.component === undefined) {
        readVal = tree.read(ROOT_REDUCER_FILE_PATH);
        declarationRecorder = tree.beginUpdate(ROOT_REDUCER_FILE_PATH);
        if (readVal) {
          let val = readVal || '';
          if (val) {
            let str = val.toString();
            if (str.indexOf('const RootReducer = (') > -1) {
              declarationRecorder.insertLeft(
                0,
                `import { ${new GeneratorHelper().reducerGenerator.bind(options)()} } from '../components/${strings.dasherize(options.name)}/${strings.dasherize(
                  options.name
                )}.reducer';\n`
              );
            }
            if (str.indexOf('combineReducers({') > -1) {
              declarationRecorder.insertRight(str.indexOf('combineReducers({') + 'combineReducers({'.length, `${new GeneratorHelper().reducerGenerator.bind(options)()}, `);
            }
          }
        }
      } else {
        const component_reducer_path = `${getComponentPath(options)}/${strings.dasherize(options.component)}.reducer.ts`;
        readVal = tree.read(component_reducer_path);
        declarationRecorder = tree.beginUpdate(component_reducer_path);
        if (readVal) {
          let val = readVal || '';
          if (val) {
            let str = val.toString();
            declarationRecorder.insertLeft(
              0,
              `import { ${new GeneratorHelper().reducerGenerator.bind(options)()} } from './${strings.dasherize(options.name)}/${strings.dasherize(options.name)}.reducer';\n`
            );
            if (str.indexOf('combineReducers({') > -1) {
              declarationRecorder.insertRight(str.indexOf('combineReducers({') + 'combineReducers({'.length, `${new GeneratorHelper().reducerGenerator.bind(options)()}, `);
            }
          }
        }
      }
      tree.commitUpdate(declarationRecorder);
    };
  }

  addDeclarationToSagas(options: any): Rule {
    return (tree: Tree) => {
      let declarationRecorder, readVal;
      if (options.component === undefined) {
        readVal = tree.read(ROOT_SAGA_FILE_PATH);
        declarationRecorder = tree.beginUpdate(ROOT_SAGA_FILE_PATH);
        if (readVal) {
          let val = readVal || '';
          if (val) {
            let str = val.toString();
            declarationRecorder.insertLeft(
              0,
              `import ${new GeneratorHelper().reducerGenerator.bind(options)()} from '../components/${strings.dasherize(options.name)}/${strings.dasherize(options.name)}.saga';\n`
            );
            if (str.indexOf('yield [') > -1) {
              declarationRecorder.insertRight(str.indexOf('yield [') + 'yield ['.length, `...${new GeneratorHelper().reducerGenerator.bind(options)()}, `);
            }
          }
        }
      } else {
        const component_saga_path = `${getComponentPath(options)}/${strings.dasherize(options.component)}.saga.ts`;
        readVal = tree.read(component_saga_path);
        declarationRecorder = tree.beginUpdate(component_saga_path);
        if (readVal) {
          let val = readVal || '';
          if (val) {
            let str = val.toString();
            declarationRecorder.insertLeft(
              0,
              `import ${new GeneratorHelper().reducerGenerator.bind(options)()} from './${strings.dasherize(options.name)}/${strings.dasherize(options.name)}.saga';\n`
            );
            if (str.indexOf('export default [') > -1) {
              declarationRecorder.insertRight(str.indexOf('export default [') + 'export default ['.length, `...${new GeneratorHelper().reducerGenerator.bind(options)()}, `);
            }
          }
        }
      }

      tree.commitUpdate(declarationRecorder);
    };
  }

  generate(options: any): Rule {
    options.path = options.path ? normalize(options.path) : options.path;

    let data = {
      ...strings,
      ...options
    };

    data = {
      ...data,
      ...{
        resolveRelativePath: new GeneratorHelper().resolveRelativePath.bind(data),
        generateFilePath: new GeneratorHelper().generateFilePath.bind(data),
        resolveApiModule: new GeneratorHelper().resolveApiModule.bind(data),
        resolveModelName: new GeneratorHelper().resolveModelName.bind(data),
        generateFilePathForApi: new GeneratorHelper().generateFilePathForApi.bind(data),
        actionGenerator: new GeneratorHelper().actionGenerator.bind(data),
        pluralizeName: new GeneratorHelper().pluralizeName.bind(data),
        reducerGenerator: new GeneratorHelper().reducerGenerator.bind(data),
        fullReducerGenerator: new GeneratorHelper().fullReducerGenerator.bind(data),
        routerGenerator: new GeneratorHelper().routerGenerator.bind(data)
      }
    };

    const items = options.with.split(',');

    const merge = [];
    const declaretion = [];

    if (items.indexOf('model') > -1) {
      merge.push(apply(url('./files/admin_user_component/app/models'), [template(data), move(`${options.sourceDir}/app/models`)]));
    }
    if (items.indexOf('api') > -1) {
      if (items.indexOf('model') === -1 && options['model'] === undefined) {
        throw new Error('model option is required to generate api if model is not generated by --with option');
      }
      merge.push(apply(url('./files/admin_user_component/app/api'), [template(data), move(getApiFolderPath(options))]));
      declaretion.push(this.addDeclarationToApiService(options));
      declaretion.push(this.addDeclarationToApiModule(options));
    }
    if (items.indexOf('component') > -1) {
      if (items.indexOf('api') === -1 && options['api'] === undefined) {
        throw new Error('api option is required to generate component if api is not generated by --with option');
      }
      merge.push(apply(url('./files/admin_user_component/app/components'), [template(data), move(getComponentPath(options))]));
      if (options.noredux !== true) {
        declaretion.push(this.addDeclarationToReducer(options));
        declaretion.push(this.addDeclarationToSagas(options));
      }
    }
    return chain([...merge.map(item => mergeWith(item)), ...declaretion]);
  }
}
